#!/usr/bin/perl -w
# $Id: Util.pm,v 1.18 2001/03/10 03:42:02 pbaltz Exp $
# author: jmayer@cobalt.com
#
# MPbug fixed.

package Sauce::Util;

use strict;
use vars qw( $DEFAULT_MODE );

$DEFAULT_MODE = 0640;

use FileHandle;

sub SAFEMODE { O_WRONLY | O_CREAT | O_EXCL };

sub escape_perl_metachars
{
    $_ =~ s/([\\\|\(\)\[\{\^\$\*\+\?\.])/\\$1/g;
    return $_;
}

sub keyvalue_edit_fcn
# a convenient routine for key=value editing
# pass in ($input, $output, $comments, $separator, $options, %settings)
# options: noappend,nosave
{
    my ($input, $output, $comments, $sep, $options, %settings) = @_;
    my $keys = ':' . join(':', keys %settings) . ':';
    my ($spaces,$key);
    
    while (<$input>) {
        # skip comments
        if ($comments and /^\s*[$comments]/o) {
            print $output $_;
            next;
        }

	# find a key=value, strip away pre- and post-spaces
	if (/^(\s*)(.+?)\s*$sep/ and not ($options =~ /\bnosave\b/)) {
	    ($spaces, $key) = ($1, $2);
	    if ($keys =~ /:$key:/) {
		print $output "$spaces$key$sep$settings{$key}\n";
		delete $settings{$key};
		next;
	    }
	}
	print $output $_;
    }
	
    # append anything that's missing
    unless ($options =~ /\bnoappend\b/) {
	foreach $key (keys %settings) {
	    print $output "$key$sep$settings{$key}\n";
	}
    }

    return 0;
}

sub hash_edit_function
# inspired by keyvalue_edit_fcn, but gives greater control over how the file is
# edited
# arguments are 
# ($input, $output, $comments, \%seperator , \%modifications_hash, $delete_not_found)
# $input and $ouput are the already opened input and output file handles
# $comments is a string with characters that when found at the beginning of a line
# 			should be skipped as comment lines (eg $comments = "#" skips lines that
#			begin with "#")
# \%seperator is a hash reference to a hash structered as follows
#			$seperator = {
#							# regular expression used to seperate a line into a
#							# key and a value 
#							# (eg. ($key, $value) = split(/$seperator->{re}/, $line))
#							're' => 'seperator regex',
#							
#							# actual text to put in when writing a key value
#							# pair to $output
#							# (eq print $output "$key$seperator->{val}$value\n"
#							'val' => 'seperator'
#						 };
# \%modifications_hash is a hash reference to a hash describing which pairs to edit
# 			in the input file
#			for example:
# 			$modifications_hash = {
#							'foo' => 'bar',
#							'bar' => ''
#								  };
#			If there was a key-value pair with key=foo, the value would be changed
#			to bar, or if the key foo doesn't exist it would be added to the file
#			with value bar.
#			If there is a key-value pair with key=bar, that line would be removed
#			from the file.
# $delete_not_found indicates whether a key found in the input file that is not found
#			in the modification hash should be removed from the file.  This defaults
#			to false, so pairs not in the modification hash are untouched.
{
	my $in = shift;
	my $out = shift;
	my $comments = shift;
	my $sep = shift;
	my $mod_hash = shift;
	my $delete_not_found = shift;

	$delete_not_found ||= 0;

	while (<$in>)
	{
		# skip comments
		if (/^\s*[$comments]/) 
		{
			print $out $_;
			next;
		}

		# need to split up the line, so we can know exactly what to look for
		my @parts = split(' ');

		# save leading whitespace, although it probably doesn't matter
		/^(\s*)$parts[0]/;
		my $space = $1;

		if (join(' ', @parts) =~ /^(\S+) $sep->{re}/)
		{
			my $key = $1;

			if (exists $mod_hash->{$key})
			{
				# indicate key deletion by an existing key with no value
				if ($mod_hash->{$key} ne "")
				{
					print $out "$space$key$sep->{val}$mod_hash->{$key}\n";
				}
				delete $mod_hash->{$key};
			}
			elsif (not $delete_not_found)
			{
				print $out $_;
			}
		}
		else
		{
			print $out $_;
		}
	}

	# add any values that didn't have a previous entry in the file
	for my $key (keys %$mod_hash)
	{
		if ($mod_hash->{$key} ne "")
		{
			print $out "$key$sep->{val}$mod_hash->{$key}\n";
		}
	}

	return 1;
}

# $ok = lockfile ($filename)
#
# we're not using flock because flock seems to be broken in Perl on mips.
sub lockfile
{
  my $filename = shift;

  my $lockfile = $filename;
  $lockfile =~ s/[^a-zA-Z0-9\-\.]+/_/g;
  $lockfile = "/var/lock/editlock_" . $lockfile;
  
  # lock the file
  for (my $i = 0; $i < 5; $i++) {
  	last if (!-e $lockfile);
    sleep(1);
  }
  unlink($lockfile);
  my $fh = new FileHandle("$lockfile", SAFEMODE);
  if ($fh) {
  	print $fh $$,"\n";
    $fh->close();
  }
  
  return 1; # ok
}

# $ok = unlockfile ($filename)
#
# we're not using flock because flock seems to be broken in Perl on mips.
sub unlockfile
{
	my $filename = shift;

  my $lockfile = $filename;
  $lockfile =~ s/[^a-zA-Z0-9\-\.]+/_/g;
  $lockfile = "/var/lock/editlock_" . $lockfile;
  
  # unlock the file
  unlink($lockfile);

  return 1; # ok
}


# $ok = editfile ($filename, $function, ...)
#
# Used to edit $filename in a safe, consistent way.
#
# $function is executed such that $_[0] is the input filehandle,
# $_[1] is the output filehandle, and $_[2...] contains any additional
# data passed into editfile.
sub editfile
{
  my $filename = shift;
  my $function = shift;

  Sauce::Util::lockfile($filename);
  
  # attempt to handle the case where the file is missing:
  if (!-e $filename) {
      print STDERR "Warning: File $filename was missing.\n";
      my $fh = new FileHandle("$filename", SAFEMODE);
    if ($fh) {
      $fh->close();
    } else {
      print STDERR "Could not create $filename: $!\n";
      return 0; # fail
    }
  }

  # move file to a backup location
  my $newfile = $filename . "~";

  # open for reading  
  my $fin = new FileHandle("<$filename");
  if (!$fin) {
    print STDERR "Could not open $filename for reading: $!\n";
    return 0;
  }
  
  # open for writing
  my @statbuf = stat($filename);
  unlink($newfile);
  my $fout = new FileHandle("$newfile", SAFEMODE, $statbuf[2]);
  if (!$fout) {
    print STDERR "Could not open $newfile for writing: $!\n";
    return 0;
  }
  
  # process file
  my $oldfh = select(); # save stdout file handle
  my $ret = &$function($fin, $fout, @_);
  select($oldfh); # restore stdout file handle
  
  # close file handles
  $fout->close();
  $fin->close();

  # set up permissions:
  copy_access_bits($filename, $newfile);
  # defined($fmode) && chmod($fmode, $newfile);

  # pull the old file switcheroo:
  ($ret =~ m/^FAIL/) || switch_files($filename, $newfile);

  # unlock the file
  Sauce::Util::unlockfile($filename);
  
  # return success
  return $ret;
}

# my $ok = process_template($templatefile, $hashref)
sub process_template
{
  my $filename = shift;
  my $hashref = shift;
  my $fmode = shift || $DEFAULT_MODE;

  my $outfile = $filename;
  $outfile =~ s/\.te?mpla?t?e?$//;
  if ($outfile eq $filename) { $outfile .= ".out"; }
  
  my $tmpfile = $outfile . "~";
    
  # lock
  Sauce::Util::lockfile($outfile);  

  # attempt to handle the case where the file is missing:
  if (!-e $filename) {
    print STDERR "Missing file: $filename\n";
    return 0;
  }

  # open for reading  
  my $fin = new FileHandle("<$filename");
  if (!$fin) {
    print STDERR "Could not open $filename for reading: $!\n";
    return 0;
  }
  
  # open for writing
  my @statbuf = stat($filename);
  unlink($tmpfile);
  my $fout = new FileHandle($tmpfile, SAFEMODE, $statbuf[2]);
  if (!$fout) {
  	print STDERR "Could not open $tmpfile for writing: $!\n";
    return 0;
  }
      
  while (defined($_ = <$fin>)) {
    s/\$\{(\S+?)\}/$hashref->{$1}/g;
    print $fout $_;
  } 

  # close file handles
  $fout->close();
  $fin->close();
  
  # set up permissions:
  if (-e $outfile) {
    copy_access_bits($outfile, $tmpfile);
  }
  defined($fmode) && chmod($fmode, $tmpfile);

  # pull the old file switcheroo:
  if (-e $outfile) {
    switch_files($outfile, $tmpfile);
  } else {
    link($tmpfile, $outfile);
  }

  # unlock
  Sauce::Util::unlockfile($outfile);  
  
  return 1;
}

# editblock( $filename, $fun, $start_line, $end_line, @args );
#
# Caveat: start_line and end_line must start at the beginning of a new line.
#
# FIXME: there is no way to add an optional "mode" arguement to this
# function w/o breaking legacy code.  :-(  All code using this function
# should be fixed.
sub editblock
{
  my $filename = shift;
  my $fun = shift;
  my $start_line = shift;
  my $end_line = shift;
  my @args = @_;
  
  my $ret = 1;

  Sauce::Util::lockfile($filename);
  
  # attempt to handle the case where the file is missing:
  if (!-e $filename) {
  	print STDERR "Warning: File $filename was missing.\n";
  	my $fh = new FileHandle($filename, SAFEMODE);
    if ($fh) {
    	$fh->close();
    } else {
    	print STDERR "Could not create $filename: $!\n";
      return 0; # fail
    }
  }

  # open for reading  
  my $fin = new FileHandle("<$filename");
  if (!$fin) {
    print STDERR "Could not open $filename for reading: $!\n";
    return 0;
  }

  # Open the three temporary files (Start, insection and endof.)
  unlink( $filename . ".start");
  my $fstart = new FileHandle("$filename.start", SAFEMODE);
  if (!$fstart) {
	  print STDERR "Could not open $filename.start for writing: $!\n";
	  return 0;
  }

  unlink( $filename . ".middle" );
  my $fmiddle = new FileHandle("$filename.middle", SAFEMODE);
  if (!$fmiddle) {
	  print STDERR "Could not open $filename.middle for writing: $!\n";
	  return 0;
  }

  unlink ( $filename . ".end" );
  my $fend = new FileHandle("$filename.end", SAFEMODE);
  if (!$fend) {
	  print STDERR "Could not open $filename.end for writing: $!\n";
	  return 0;
  }

  # precompile regexi:
  my $start_re = qr/^$start_line\s*$/;
  my $end_re = qr/^$end_line\s*$/;
  
  # split file into three sections
  my $mode = 'start';
  while (defined($_ = <$fin>)) {
  	if ($mode eq 'start') {
    	# initial state
    	if ($_ =~ $start_re) {
      	# encountered start tag
      	$mode = 'middle';
	# We don;t print it so step on.
	next;
      }
      print $fstart $_;
    } elsif ($mode eq 'middle') {
      # inside block state 
      if ($_ =~ $end_re) {
      	# encountered an end tag
      	$mode = 'end';
	next;
      }
      print $fmiddle $_;
    } else {
      print $fend $_;
    }
  } # end of while loop

  # close input handle;
  $fin->close();
  # Close our three section handler;
  $fstart->close();
  $fmiddle->close();
  $fend->close();

  # Now we call editfile on just the section that the user wanted.
  $ret = editfile("$filename.middle",$fun,@args);

  unless( $fstart->open("<$filename.start") &&
	  $fmiddle->open("<$filename.middle") &&
	  $fend->open("<$filename.end") ) 
  {
    print STDERR "Could not reopen all files for $filename: $!\n";
    return 0;
  }

  # backup file name
  my $backupfile = $filename . '~';

  # copy to a backup file now
  unlink($backupfile);
  my $fout = new FileHandle($backupfile, SAFEMODE, 0600);
  $fout->print( join('',<$fstart> ) );
  $fout->print( $start_line . "\n" );
  $fout->print( join('',<$fmiddle> ) );
  $fout->print( $end_line . "\n" );
  $fout->print( join('',<$fend> ) );

  # Close out main file.
  $fout->close;

  # now, swap backup and new files
  switch_files($filename, $backupfile);

  # Close all out sub files.
  $fstart->close();
  $fmiddle->close();
  $fend->close();

  # Clean up out temp files.

  #unlink("$filename.start");
  #unlink("$filename.middle");
  #unlink("$filename.end");

  # unlock the file
  Sauce::Util::unlockfile($filename);
  
  # return success
  return $ret;
}

# replaceblock( $filename, $start_line, $data, $end_line, $filemode );
#
# replaces the section of $filename denoted by the $start_line and
# $end_line with the contents of $data.
#
# Caveat: if $data is an empty string, the (empty) block is left in
# place in the file as a placeholder.  OTOH, if $data is undefined,
# the block (including start and stop tags) are completely removed
# from the file.
#
# Caveat: start_line and end_line must start at the beginning of a new line.
sub replaceblock
{
  my $filename = shift;
  my $start_line = shift;
  my $data = shift;
  my $end_line = shift;
  my $fmode = shift || undef;
  
  my $ret = 1;

  Sauce::Util::lockfile($filename);
  
  # attempt to handle the case where the file is missing:
  if (!-e $filename) {
    print STDERR "Warning: File $filename was missing.\n";
    unlink($filename);
    my $fh = new FileHandle("$filename", SAFEMODE);
    if ($fh) {
    	$fh->close();
    } else {
      print STDERR "Could not create $filename: $!\n";
      return 0; # fail
    }
  }

  # make a place for the new edition:
  my $newfile = $filename . '~';
  unlink($newfile) if (-e $newfile);

  # open for reading  
  my $fin = new FileHandle("<$filename");
  if (!$fin) {
    print STDERR "Could not open $filename for reading: $!\n";
    return 0;
  }


  # open for writing
  my @statbuf = stat($filename);
  unlink($newfile);
  my $fout = new FileHandle("$newfile", SAFEMODE, $statbuf[2]);
  if (!$fout) {
    print STDERR "Could not open $newfile for writing: $!\n";
    return 0;
  }

   # precompile regexi:
  my $start_re = qr/^$start_line\s*$/;
  my $end_re = qr/^$end_line\s*$/;
  
 
  # process file:
  my $mode = 'init';
  while (defined($_ = <$fin>)) {
    if ($mode eq 'init') {
      # initial state (print anything but a start tag)
      if ($_ =~ $start_re) {
        # encountered start tag
        $mode = 'inside';
        if (defined($data)) {
          print $fout $start_line,"\n",$data,"\n",$end_line,"\n";
        }
        next;
      }
      print $fout $_;
      next;
    }
    if ($mode eq 'inside') {
      # inside block state (don't print anything)
      if ($_ =~ $end_re) {
        # encountered an end tag
        $mode = 'outside';
      }
      next;
    }
    if ($mode eq 'outside') {
      # past block state (print anything)
      print $fout $_;
      next;
    }
  } # end of while loop
  if ($mode eq 'init' && defined($data)) {
    # tag was never found, so let's make one:
    print $fout $start_line,"\n",$data,"\n",$end_line,"\n";
  }
    
  # close file handles
  $fout->close();
  $fin->close();
  
  # set up permissions:
  copy_access_bits($filename, $newfile);
  defined($fmode) && chmod($fmode, $newfile);

  # pull the old file switcheroo:
  switch_files($filename, $newfile);

  # unlock the file
  Sauce::Util::unlockfile($filename);
  
  # return success
  return $ret;
}

sub switch_files
{
  my ($f1, $f2) = (shift, shift);

  if (!-e $f1) { return -9; }
  if (!-e $f2) {
    link $f1, $f2;
    return 1;
  }
  
  my $tmp = $f1 . ".$$.tmp";

  do { link ($f1, $tmp) } || return -1;
  unlink($f1) || do {
    # undo
    unlink $tmp;
    return -2;
  };
  
  link ($f2, $f1) && unlink($f2) || do {
    # undo
    unlink($f1);
    link ($tmp, $f1) && unlink($tmp);
    return -3;
  };
  
  link ($tmp, $f2) || do {
    # undo
    link($f1, $f2);
    unlink($f1);
    link ($tmp, $f1) && unlink($tmp);
    return -4;
  };
  unlink($tmp);
  return 1;
} 
  

sub copy_access_bits
{
  my ($srcfile, $destfile) = (shift, shift);

  my @buf;
  @buf = stat($srcfile);
  if (!@buf) {
    print STDERR "Couldn't stat $srcfile\n";
    return 0;
  }
  
  my $mode = $buf[2] & 07777;
  chmod($mode, $destfile) || do {
    print STDERR "Couldn't chmod $mode, $destfile\n";
    return 0;
  };
  
  my $uid = $buf[4];
  my $gid = $buf[5];
  chown ($uid, $gid, $destfile) || do {
    print STDERR "Couldn't chown $destfile\n";
    return 0;
  };
  
  return 1;
}

sub force_access_permissions
{
  my $file = shift;
  my $uid = shift;
  my $gid = shift;
  my $mode = shift;
  
  my @buf = stat($file);
  if (!@buf) {
    print STDERR "$0: file \"$file\" unexpectedly missing.\n";
    return 0;    
  }
  
  if (!int($uid)) {
    my @pwent = getpwnam($uid);
    if (!@pwent) {
      print STDERR "$0: user \"$uid\" does not exist.\n";
      return 0;
    }
    $uid = $pwent[2];
  }

  if (!int($gid)) {
    my @ent = getgrnam($gid);
    if (!@ent) {
      print STDERR "$0: group \"$gid\" does not exist.\n";
      return 0;
    }
    $gid = $ent[2];
  }
  
  {
    my $flag = 0;
    if ($buf[4] != $uid) {
      print STDERR "$0: file \"$file\" has wrong uid($buf[4]),",
      	" correcting to uid($uid).\n";
      $flag = 1;
    }
    if ($buf[5] != $gid) {
      print STDERR "$0: file \"$file\" has wrong gid($buf[5]),",
      	" correcting to gid($gid).\n";
      $flag = 1;
    }
    if ($flag) {
      if (!(chown $uid, $gid, $file)) {
      	print STDERR "$0: Could not chown file \"$file\".\n";
      }
    }
  };
  
  if ($buf[2] != $mode) {
    print STDERR "$0: file \"$file\" had incorrect mode 0",oct($buf[2]),"\n";
  }
}

sub replace_unique_entries
{
  my ($fin, $fout) = (shift, shift);
  my $oid = shift;
  my $hash = shift;
  my $inside = 0;
  my $startline = "# Object ID $oid owns these lines:";
  my $stopline ="# End of Object ID $oid section";
  
  my $errs = 0;
  
  while (defined($_ = <$fin>)) {
    if ($inside != 1) {
      print $fout $_;
      if ((!m/^\s*#/) && (m/^\s*(\S+?):\s+(.*)/)) {
      	# check uniqueness
	if (defined($hash->{$1})) {
	  $errs++;
	}
      }
      if (m/^$startline/) { 
      	# dump contents
	foreach my $k (keys %$hash) {
	  print $fout "$k:\t",$hash->{$k},"\n";
	}
      	$inside = 1; 
      }
      next;
    } else {
      if (m/^$stopline/) { print $fout $_; $inside = 3; }
      next;
    }
  }
  
  if ($inside == 0) {
    print $fout $startline,"\n";
    foreach my $k (keys %$hash) {
      print $fout "$k:\t",$hash->{$k},"\n";
    }
    print $fout $stopline,"\n";
  }
  
  if ($errs) { return "FAIL"; }
  return 1;
}



1;
